#ifndef __IDROPTARGET_CPP__
#define __IDROPTARGET_CPP__


//Members
CDropTarget::CDropTarget( bool delself, CDropTarget** p )
{
#ifdef _DEBUG
	printf("IDropTarget::IDropTarget\n");
	if( !p ) printf("IDropTarget constructed without pthis\n");
#endif //_DEBUG
	m_cRefCount = 0;
	deleteself = delself;
	pthis = p;
}

CDropTarget::~CDropTarget()
{
#ifdef _DEBUG
	printf("IDropTarget::~IDropTarget\n");
	if( m_cRefCount != 0 )
		printf("Deleting %s too early 0x%x.m_cRefCount = %d\n", "IDropTarget", this, m_cRefCount);
#endif //_DEBUG
	if( pthis ) *pthis = NULL;
}


//IUnknown methods
STDMETHODIMP CDropTarget::QueryInterface( REFIID iid, void** ppvObject )
{
#ifdef _DEBUG
	printf("IDropTarget::QueryInterface\n");
#endif //_DEBUG

	*ppvObject = NULL;

	if ( iid == IID_IUnknown ) *ppvObject = (IUnknown*)this;
	else if ( iid == IID_IDropTarget ) *ppvObject = (IDropTarget*)this;

	if(*ppvObject){
		((IUnknown*)*ppvObject)->AddRef();
		return S_OK;
	}

	return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) CDropTarget::AddRef( void )
{
#ifdef _DEBUG
	printf("IDropTarget::AddRef\n");
#endif //_DEBUG
	return ++m_cRefCount;
}

STDMETHODIMP_(ULONG) CDropTarget::Release( void )
{
#ifdef _DEBUG
	printf("IDropTarget::Release\n");
#endif //_DEBUG
	if( --m_cRefCount == 0 && deleteself ) delete this;
	return m_cRefCount;
}


//IDropTarget methods
STDMETHODIMP CDropTarget::DragEnter( IDataObject* pDataObject, DWORD grfKeyState, POINTL pt, DWORD* pdwEffect )
{
#ifdef _DEBUG
	printf("IDropTarget::DragEnter\n");
#endif //_DEBUG

	pDataObject->AddRef();

	CreateCaret( hexwnd.hwnd, (HBITMAP)1, 2, hexwnd.cyChar );

	return DragOver( grfKeyState, pt, pdwEffect );
}

STDMETHODIMP CDropTarget::DragOver( DWORD grfKeyState, POINTL pt, DWORD* pdwEffect )
{
#ifdef _DEBUG
	printf("IDropTarget::DragOver\n");
#endif //_DEBUG

	LastKeyState = grfKeyState;

	DWORD dwOKEffects = *pdwEffect;

	//We only move or copy if we are not read-only
	//And only if the source supports it
	//but in read-only mode we accept drag-drop loading (CF_HDROP)
	//this is handled by OLE it seems if the main window DragAcceptFiles( hwnd, TRUE )
	if( hexwnd.bReadOnly ){
		*pdwEffect = DROPEFFECT_NONE;
	}
	else if( grfKeyState & MK_CONTROL )
		*pdwEffect = dwOKEffects & DROPEFFECT_COPY ? DROPEFFECT_COPY : ( dwOKEffects & DROPEFFECT_MOVE ? DROPEFFECT_MOVE : DROPEFFECT_NONE );
	else
		*pdwEffect = dwOKEffects & DROPEFFECT_MOVE ? DROPEFFECT_MOVE : ( dwOKEffects & DROPEFFECT_COPY ? DROPEFFECT_COPY : DROPEFFECT_NONE );

	POINT p = {pt.x, pt.y};
	ScreenToClient(hexwnd.hwnd,&p);
	hexwnd.iMouseX = p.x;
	hexwnd.iMouseY = p.y;

	if( *pdwEffect == DROPEFFECT_NONE ) HideCaret( hexwnd.hwnd );
	else hexwnd.set_drag_caret( p.x, p.y, *pdwEffect == DROPEFFECT_COPY );

	if((( p.y>(hexwnd.cyBuffer-1)*hexwnd.cyChar && hexwnd.iCurLine < hexwnd.iNumlines-hexwnd.cyBuffer)
		|| ( p.y<hexwnd.cyChar && hexwnd.iCurLine > 0 ) ) ||
		( ( p.x>hexwnd.cxBuffer*hexwnd.cxChar && hexwnd.iHscrollPos < hexwnd.iCharsPerLine-hexwnd.cxBuffer ) 
		|| ( p.x<hexwnd.cxChar && hexwnd.iHscrollPos > 0)))
	{
		if (hexwnd.bScrollTimerSet==FALSE)
		{
			// Set timer for scrolling.
			SetTimer (hexwnd.hwnd, SCROLLTIMERID, SCROLLTIMERDURATION, NULL);
			hexwnd.bScrollTimerSet = TRUE;
			*pdwEffect |= DROPEFFECT_SCROLL;
		}
	}
	
	hexwnd.dragging = true;

	return S_OK;
}

STDMETHODIMP CDropTarget::DragLeave ( void )
{
#ifdef _DEBUG
	printf("IDropTarget::DragLeave\n");
#endif //_DEBUG

	CreateCaret( hexwnd.hwnd, NULL, hexwnd.cxChar, hexwnd.cyChar );

	KillTimer( hexwnd.hwnd, SCROLLTIMERID ); hexwnd.bScrollTimerSet = FALSE;
	
	hexwnd.dragging = true;

	return S_OK;
}

STDMETHODIMP CDropTarget::Drop( IDataObject* pDataObject, DWORD grfKeyState, POINTL pt, DWORD* pdwEffect )
{
#ifdef _DEBUG
	printf("IDropTarget::Drop\n");
#endif //_DEBUG

	{
		DWORD temp = LastKeyState;
		//Get the position of the drop
		DragOver( grfKeyState, pt, pdwEffect );
		LastKeyState = temp;
	}
	//Remove the effects
	DragLeave();

	bool copying = *pdwEffect & DROPEFFECT_COPY ? true : false ;

	hexwnd.bDroppedHere = TRUE;

	if( hexwnd.bMoving ){
		iMove1stEnd = hexwnd.iStartOfSelection;
		iMove2ndEndorLen = hexwnd.iEndOfSelection;
		if( iMove1stEnd > iMove2ndEndorLen ) swap( iMove1stEnd, iMove2ndEndorLen );
		if( !copying && hexwnd.new_pos > iMove2ndEndorLen ) hexwnd.new_pos += iMove1stEnd - iMove2ndEndorLen - 1;
		iMovePos = hexwnd.new_pos;
		iMoveOpTyp = copying ? OPTYP_COPY : OPTYP_MOVE;
		if( grfKeyState&MK_SHIFT ){
			int len = iMove2ndEndorLen - iMove1stEnd + 1;
			if( copying ){
				//just realloc & memmove
				int upper = 1+hexwnd.DataArray.GetUpperBound();
				if( iMovePos+len > upper ){
					if( hexwnd.DataArray.SetSize( iMovePos+len ) ){
						hexwnd.DataArray.ExpandToSize();
						memmove(&hexwnd.DataArray[iMovePos],&hexwnd.DataArray[iMove1stEnd],len);
					}
				} else {
					memmove(&hexwnd.DataArray[iMovePos],&hexwnd.DataArray[iMove1stEnd],len);
				}
			} else {/*moving*/
				if(iMovePos>iMove1stEnd){/*up*/
					hexwnd.CMD_move_copy( 1, 0 );
					hexwnd.DataArray.RemoveAt(iMovePos+len,len);
				} else {/*down*/
					memmove(&hexwnd.DataArray[iMovePos],&hexwnd.DataArray[iMove1stEnd],len);
					hexwnd.DataArray.RemoveAt((iMovePos-iMove1stEnd>=len?iMove1stEnd:iMovePos+len),len);
				}
			}
			hexwnd.iStartOfSelection = iMovePos;
			hexwnd.iEndOfSelection = iMovePos+len-1;
			hexwnd.m_iFileChanged = hexwnd.bFilestatusChanged = hexwnd.bSelected = TRUE;
			hexwnd.update_for_new_datasize();
		} else hexwnd.CMD_move_copy( 1 );
	} else {

		HRESULT err = E_UNEXPECTED;

		//Get all the data formats
		IEnumFORMATETC* iefe = NULL;
		pDataObject->EnumFormatEtc( DATADIR_GET, &iefe );
		if( !iefe ){
#ifdef _DEBUG
			printf("Unable to create a drag-drop data enumerator\n");
#endif //_DEBUG
			goto ERR;
		}
		try{
			iefe->Reset();
		}
		catch(...){
			goto ERR;
		}

		FORMATETC* fel;fel = NULL;
		UINT numfe;numfe = 0;
		for(;;){
			void* temp = realloc( fel, ( numfe + 1 ) * sizeof(FORMATETC) );
			if( temp != NULL ){
				fel = (FORMATETC*) temp;
				temp = NULL;
				int r;
				r = iefe->Next( 1, &fel[numfe], NULL);
				if( r != S_OK ) break;//No more formats
				numfe++;
			} else {
#ifdef _DEBUG
				printf("Not enough memory for the drag-drop format list\n");
#endif //_DEBUG
				goto ERR_ENUM;
			}
		}

		/*If there is just one format or "BinaryData" is present (and the user
		requested to use it whenever it is present) we don't need feedback from
		the user*/
		bool NeedToChooseFormat;NeedToChooseFormat = false;
		bool NeedToInsertData;NeedToInsertData = false;
		UINT IndexOfDataToInsert;IndexOfDataToInsert = 0;
		if( numfe == 0 ){
			MessageBox( hexwnd.hwnd, "No data to insert", "Drag-drop", MB_OK );
		} else if( numfe == 1 ){
			NeedToInsertData = true;
		} else {
			bool BinaryDataIsPresent = false;
			UINT IndexOfBinaryData = 0;
			if( hexwnd.UseBinaryDataIfPresent ){
				for( UINT i=0; i < numfe; i++ ){
					if( fel[i].cfFormat == CF_BINARYDATA ){
						BinaryDataIsPresent = true;
						IndexOfBinaryData = i;
						break;
					}
				}
			}
			NeedToInsertData = true;
			if( BinaryDataIsPresent == false ) NeedToChooseFormat = true;
		}

		bool NeedToChooseMoveorCopy;
		NeedToChooseMoveorCopy = ( (LastKeyState|grfKeyState) & ( MK_MBUTTON | MK_RBUTTON ) ) ? true : false;
		if( NeedToChooseFormat == false && NeedToChooseMoveorCopy == true ){
			HMENU hm = CreatePopupMenu();
			InsertMenu( hm, 0, MF_BYPOSITION|MF_STRING, 1, "Move" );
			InsertMenu( hm, 1, MF_BYPOSITION|MF_STRING, 2, "Copy" );
			InsertMenu( hm, 2, MF_BYPOSITION|MF_SEPARATOR, 0, 0 );
			InsertMenu( hm, 3, MF_BYPOSITION|MF_STRING, 0, "Cancel" );
			BOOL mi = TrackPopupMenuEx( hm, TPM_NONOTIFY|TPM_RIGHTBUTTON|TPM_RETURNCMD, pt.x, pt.y, hexwnd.hwnd, NULL );
			DestroyMenu( hm );
			if( mi == 0 ){ err = S_OK; *pdwEffect = DROPEFFECT_NONE; goto ERR_ENUM; }
			copying = ( mi == 1 ? false : true );
		} else if( NeedToChooseFormat ) {
			DROPPARAMS params;
			params.needeffect = NeedToChooseMoveorCopy;
			params.effect = copying;
			params.formats = fel;
			params.numformats = numfe;
			params.formatnumber = IndexOfDataToInsert;
			int ret = DialogBoxParam( hexwnd.hInstance, MAKEINTRESOURCE(IDD_DRAGDROP), hexwnd.hwnd, (DLGPROC) DragDropDlgProc, (LPARAM)&params );
			if( ret < 0 ){ err = S_OK; *pdwEffect = DROPEFFECT_NONE; goto ERR_ENUM; }//An error occured or the user canceled the operation
			copying = params.effect;IndexOfDataToInsert = ret;
		}

		/*
		E_INVALIDARG
		E_UNEXPECTED
		E_OUTOFMEMORY
		S_OK
		DV_E_LINDEX
		DV_E_FORMATETC
		DV_E_TYMED
		DV_E_DVASPECT
		OLE_E_NOTRUNNING
		STG_E_MEDIUMFULL
		*/

		FORMATETC fe;
		STGMEDIUM stm;
		int ret;
		fe = fel[IndexOfDataToInsert];
		/*It is important that when debugging (with M$VC at least) you do not step __into__ the below GetData call
		  If you do the app providing the data source will die & GetData will return OLE_E_NOTRUNNING or E_FAIL
		  The solution is to step over the call*/
		if ( (ret = pDataObject->GetData( &fe, &stm )) == S_OK )
		{
			//Get len
			size_t len = 0;
			switch( stm.tymed ){
				case TYMED_HGLOBAL: len = GlobalSize( stm.hGlobal ); break;
				case TYMED_FILE:{
					int fh = _wopen( stm.lpszFileName, _O_BINARY | _O_RDONLY );
					if( fh != -1 ){
						len = _filelength( fh );
						if( len == (size_t)-1 ) len = 0;
						_close( fh );
					}
				} break;
				case TYMED_ISTREAM:{
					STATSTG stat;
					if( S_OK == stm.pstm->Stat( &stat, STATFLAG_NONAME ) ) len = (size_t)stat.cbSize.LowPart;
				} break;
				//This case is going to be a bitch to implement so it can wait for a while
				//It will need to be a recursive method that stores the STATSTG structures (+ the name), contents/the bytes of data in streams/property sets
				case TYMED_ISTORAGE:{ MessageBox( hexwnd.hwnd, "TYMED_ISTORAGE is not yet supported for drag-drop.\nPlease don't hesitate to write a patch & send in a diff.", "Drag-drop", MB_OK ); } break;//IStorage*
				case TYMED_GDI:{
					len = GetObject( stm.hBitmap, 0, NULL );
					if( len ){
						DIBSECTION t;
						GetObject( stm.hBitmap, len, &t );
						len += t.dsBm.bmHeight*t.dsBm.bmWidthBytes*t.dsBm.bmPlanes;
					}
				} break;//HBITMAP
				case TYMED_MFPICT:{
					len = GlobalSize( stm.hMetaFilePict );
					METAFILEPICT*pMFP=(METAFILEPICT*)GlobalLock( stm.hMetaFilePict );
					if( pMFP ){
						len += GetMetaFileBitsEx( pMFP->hMF, 0, NULL );
						GlobalUnlock( stm.hMetaFilePict );
					}
				} break;//HMETAFILE
				case TYMED_ENHMF:{
					len = GetEnhMetaFileHeader( stm.hEnhMetaFile, 0, NULL );
					DWORD n = GetEnhMetaFileDescriptionW( stm.hEnhMetaFile, 0, NULL );
					if( n && n != GDI_ERROR ) len += sizeof(WCHAR)*n;
					len += GetEnhMetaFileBits( stm.hEnhMetaFile, 0, NULL );
					n = GetEnhMetaFilePaletteEntries( stm.hEnhMetaFile, 0, NULL );
					if( n && n != GDI_ERROR ) len += sizeof(LOGPALETTE)+(n-1)*sizeof(PALETTEENTRY);
				} break;//HENHMETAFILE
				//case TYMED_NULL:break;
			}
			if( !len ){ err = S_OK; *pdwEffect = DROPEFFECT_NONE; goto ERR_ENUM; }

			//Malloc
			BYTE* data = (BYTE*)malloc(len);
			if( !data ) { err = S_OK; *pdwEffect = DROPEFFECT_NONE; goto ERR_ENUM; }
			memset( data, 0, len );

			//Get data
			bool gotdata = false;
			switch( stm.tymed ){
				case TYMED_HGLOBAL:{
					LPVOID pmem = GlobalLock( stm.hGlobal );
					if( pmem ){
						try{ memcpy( data, pmem, len ); }
						catch(...){}
						gotdata = true;
						GlobalUnlock( stm.hGlobal );
					}
				} break;
				case TYMED_FILE:{
					int fh = _wopen( stm.lpszFileName, _O_BINARY | _O_RDONLY );
					if( fh != -1 ){
						if( 0 < _read( fh, data, len ) ) gotdata = true;
						_close( fh );
					}
				} break;
				case TYMED_ISTREAM:{
					LARGE_INTEGER zero = { 0 };
					ULARGE_INTEGER pos;
					if( S_OK == stm.pstm->Seek( zero, STREAM_SEEK_CUR, &pos ) ){
						stm.pstm->Seek( zero, STREAM_SEEK_SET, NULL );
						if( S_OK == stm.pstm->Read( data, len, NULL ) ) gotdata = true;
						stm.pstm->Seek( *(LARGE_INTEGER*)&pos, STREAM_SEEK_SET, NULL );
					}
				} break;
				//This case is going to be a bitch to implement so it can wait for a while
				//It will need to be a recursive method that stores the STATSTG structures (+ the name), contents/the bytes of data in streams/property sets
				//case TYMED_ISTORAGE:{ MessageBox( hexwnd.hwnd, "TYMED_ISTORAGE is not yet supported for drag-drop.\nPlease don't hesitate to write a patch & send in a diff.", "Drag-drop", MB_OK );} break;//IStorage*
				case TYMED_GDI:{
					int l = GetObject( stm.hBitmap, len, data );
					if( l ){
						BITMAP* bm = (BITMAP*)data;
						if( bm->bmBits ) memcpy( &data[l], bm->bmBits, len-l );
						else GetBitmapBits( stm.hBitmap, len-l, &data[l] );
						gotdata = true;
					}
				} break;//HBITMAP
				case TYMED_MFPICT:{
					METAFILEPICT*pMFP=(METAFILEPICT*)GlobalLock( stm.hMetaFilePict );
					if( pMFP ){
						try{
							memcpy( data, pMFP, sizeof(*pMFP) );
							GetMetaFileBitsEx( pMFP->hMF, len - sizeof(*pMFP), &data[sizeof(*pMFP)] );
							GlobalUnlock( stm.hMetaFilePict );
							gotdata = true;
						}
						catch(...){}
					}
				} break;//HMETAFILE
				case TYMED_ENHMF:{
					DWORD i = 0, n = 0, l = len;
					n = GetEnhMetaFileHeader( stm.hEnhMetaFile, l, (ENHMETAHEADER*)&data[i] );
					l -= n; i += n;
					n = GetEnhMetaFileDescriptionW( stm.hEnhMetaFile, l/sizeof(WCHAR), (LPWSTR)&data[i] );
					if( n && n != GDI_ERROR ){
						n *= sizeof(WCHAR);l -= n; i += n;
					}
					n = GetEnhMetaFileBits( stm.hEnhMetaFile, l, &data[i] );
					l -= n; i += n;
					n = GetEnhMetaFilePaletteEntries( stm.hEnhMetaFile, NULL, 0 );
					if( n && n != GDI_ERROR ){
						LOGPALETTE* lp = (LOGPALETTE*)&data[i];
						lp->palVersion = 0x300;
						lp->palNumEntries = (USHORT)n;
						l -=sizeof(lp->palVersion)+sizeof(lp->palNumEntries);
						n = GetEnhMetaFilePaletteEntries( stm.hEnhMetaFile, l/sizeof(PALETTEENTRY), &lp->palPalEntry[0] );
						i += n*sizeof(PALETTEENTRY);
					}
					if( i ) gotdata = true;
				} break;//HENHMETAFILE
				//case TYMED_NULL:break;
			}

			ReleaseStgMedium( &stm );

			BYTE* DataToInsert = data;
			if( fe.cfFormat == CF_BINARYDATA ){
				len = *(DWORD*)data;
				DataToInsert += 4;
			} else if( fe.cfFormat == CF_TEXT || fe.cfFormat == CF_OEMTEXT ){
				try { len = strlen( (char*)data ); } catch(...){}
			} else if( fe.cfFormat == CF_UNICODETEXT ){
				try { len = sizeof(wchar_t)*wcslen( (wchar_t*)data ); } catch(...){}
			}

			//Insert/overwrite data into DataArray
			if( grfKeyState&MK_SHIFT ){
				//Overwite
				DWORD upper = 1+hexwnd.DataArray.GetUpperBound();
				if( hexwnd.new_pos+len > upper ){
					if( hexwnd.DataArray.SetSize( hexwnd.new_pos+len ) ){
						hexwnd.DataArray.ExpandToSize();
						memcpy(&hexwnd.DataArray[hexwnd.new_pos],DataToInsert,len);
						gotdata = true;
					}
				} else {
					memcpy(&hexwnd.DataArray[hexwnd.new_pos],DataToInsert,len);
					gotdata = true;
				}
			} else gotdata = ( 0 != hexwnd.DataArray.InsertAtGrow( hexwnd.new_pos, DataToInsert, 0, len ) );
			
			if(gotdata){
				hexwnd.iStartOfSelection = hexwnd.new_pos;
				hexwnd.iEndOfSelection = hexwnd.new_pos + len - 1;
				hexwnd.m_iFileChanged = hexwnd.bFilestatusChanged = hexwnd.bSelected = TRUE;
				hexwnd.update_for_new_datasize();
			}

			//Release the data
			free( data ); data = NULL;
		}

		*pdwEffect = copying  ? DROPEFFECT_COPY : DROPEFFECT_MOVE;

		err = S_OK;

ERR_ENUM:
		iefe->Release();
		free( fel );
ERR:
		pDataObject->Release();
		return err;
	}
	pDataObject->Release();

	return S_OK;
}


#endif //__IDROPTARGET_CPP__