//============================================================================================
// frhed - free hex editor

#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <direct.h>
#include <string.h>
#include <stdio.h>
#include <commctrl.h>
#include <objbase.h>
#include <shlobj.h>
#include <Shlwapi.h>
#include <limits.h>
#include <malloc.h>
#include <ctype.h>
#include <tchar.h>

//For cygwin - until oleidl.h from mingw/w32api defines them
#ifndef DD_DEFDRAGDELAY
#define DD_DEFDRAGDELAY ( 200 )
#endif
#ifndef DD_DEFDRAGMINDIST
#define DD_DEFDRAGMINDIST ( 2 )
#endif

HINSTANCE hMainInstance;
LRESULT CALLBACK MainWndProc (HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK HexWndProc (HWND, UINT, WPARAM, LPARAM);

#define Zero(a) memset(&(a),0,sizeof(a))
const UINT CF_BINARYDATA = RegisterClipboardFormat("BinaryData");

#include "resource.h"
#include "simparr.cpp"
#define OLEDD
#ifdef OLEDD
	/*In the following headers
		ULONG m_cRefCount; //The reference count that all objects based on IUnknown must have
		//The following have to do with the automatic destruction IUnknown::Release is supposed to do
		bool deleteself; //Should we delete ourself on zero reference count
		C<SomeObject>** pthis; //A pointer to a pointer that we should set to NULL on destruction*/
	#include "idt.h"
	#include "ids.h"
	#include "ido.h"
#endif //OLEDD
#include "hexwnd.h"
#include "BinTrans.cpp"
#include "diagbox.cpp"
#include "hexwnd.cpp"
#ifdef OLEDD
	#include "idt.cpp"
	#include "ids.cpp"
	#include "ido.cpp"
#endif //OLEDD

//--------------------------------------------------------------------------------------------
// WinMain: the starting point.
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, char* szCmdLine, int iCmdShow)
{

//Pabs inserted
#ifdef _DEBUG
	MessageBox(NULL,"This program has been built in development/debug mode.\nTo disable this message please obtain a release version or disable line 32 of \"main.cpp\"","frhed",MB_OK);
#endif
//end

	hMainInstance = hInstance;

	// Register window class and open window.
	HACCEL hAccel;

	MSG msg;
	WNDCLASSEX wndclass;

	ZeroMemory(&wndclass,sizeof(wndclass));
	wndclass.cbSize = sizeof(wndclass);
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = HexWndProc;
	wndclass.hInstance = hInstance;
	wndclass.hCursor = NULL;
	wndclass.lpszClassName = szHexClass;

	RegisterClassEx (&wndclass);

//Register the main window class

	wndclass.cbSize = sizeof (wndclass);
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = MainWndProc;
	wndclass.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_ICON1));
	wndclass.hCursor = LoadCursor( NULL, IDC_ARROW );
	wndclass.lpszMenuName = MAKEINTRESOURCE (IDR_MAINMENU);
	wndclass.lpszClassName = szMainClass;
//Pabs changed        // cast required to compile in mscv++6-stricter compliance with ansi
	wndclass.hIconSm =(HICON) LoadImage (hInstance, MAKEINTRESOURCE (IDI_ICON1), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
//end

	RegisterClassEx (&wndclass);

	OleInitialize(NULL);

	hWndMain = CreateWindow (szMainClass,
		"frhed window",
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
		hexwnd.iWindowX,
		hexwnd.iWindowY,
		hexwnd.iWindowWidth,
		hexwnd.iWindowHeight,
		NULL,
		NULL,
		hInstance,
		NULL);

	hHexWnd = CreateWindowEx(WS_EX_CLIENTEDGE, szHexClass,
		NULL,
		WS_TABSTOP | WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_HSCROLL,
		10,
		10,
		100,
		100,
		hWndMain,
		NULL,
		hInstance,
		NULL);

	if( hexwnd.iWindowX != CW_USEDEFAULT ){
		WINDOWPLACEMENT wp;
		wp.length = sizeof(wp);
		GetWindowPlacement( hWndMain, &wp );
		wp.showCmd = hexwnd.iWindowShowCmd;
		wp.rcNormalPosition.left = hexwnd.iWindowX;
		wp.rcNormalPosition.top = hexwnd.iWindowY;
		wp.rcNormalPosition.right = hexwnd.iWindowWidth + hexwnd.iWindowX;
		wp.rcNormalPosition.bottom = hexwnd.iWindowHeight + hexwnd.iWindowY;
		SetWindowPlacement( hWndMain, &wp );
	}
	
	ShowWindow( hWndMain, hexwnd.iWindowShowCmd );
	UpdateWindow (hWndMain);

	if( szCmdLine != NULL && strlen( szCmdLine ) != 0 )
	{
		// Command line not empty: open a file on startup.

		// BoW Patch: Remove any " by filtering the command line
		char sz[MAX_PATH];
		char* p = szCmdLine;
		strncpy( sz, szCmdLine, sizeof( sz ) );
		for( int i = 0; i < MAX_PATH; ++i )
		{
			char c = sz[i];
			if( c == 0 )
			{
				*p = 0;
				break;
			}
			else if( c != '"' )
			{
				*p++ = c;
			}
		}
		// end of patch

		char lpszPath[MAX_PATH];
		HRESULT hres = ResolveIt( hWndMain, szCmdLine, lpszPath );
		if( SUCCEEDED( hres ) )
		{
			// Trying to open a link file: decision by user required.
			int ret = MessageBox( hWndMain,
				"You are trying to open a link file.\n"
				"Click on Yes if you want to open the file linked to,\n"
				"or click on No if you want to open the link file itself.\n"
				"Choose Cancel if you want to abort opening.",
				"frhed", MB_YESNOCANCEL | MB_ICONQUESTION );
			switch( ret )
			{
			case IDYES:
				hexwnd.load_file( lpszPath );
				break;
			case IDNO:
				hexwnd.load_file( szCmdLine );
				break;
			case IDCANCEL:
				break;
			}
		}
		else
		{
			hexwnd.load_file( szCmdLine );
		}
	}

	hAccel = LoadAccelerators (hInstance, MAKEINTRESOURCE (IDR_ACCELERATOR1));

	while (GetMessage (&msg, NULL, 0, 0))
	{//Pabs inserted the modeless dialog stuff
		if ( ( hwndModelessDlg == (HWND) NULL || !IsDialogMessage(hwndModelessDlg, &msg) ) && !TranslateAccelerator (hWndMain, hAccel, &msg) )
		{//end
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}
	}

	OleUninitialize();

	return msg.wParam;
}

//--------------------------------------------------------------------------------------------
// The main window procedure.
LRESULT CALLBACK MainWndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam )
{
	switch (iMsg) {
		case WM_CREATE:
			InitCommonControls ();
			hwndStatusBar = CreateStatusWindow (
			CCS_BOTTOM | WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP,
			"Ready", hwnd, 2);
		break;
		case WM_COMMAND: return SendMessage(hHexWnd, iMsg, wParam, lParam);
		case WM_SETFOCUS: SetFocus(hHexWnd); break;
		case WM_CLOSE: return SendMessage(hHexWnd, iMsg, wParam, lParam);
		case WM_INITMENUPOPUP: hexwnd.initmenupopup( wParam, lParam ); return 0;
		case WM_DROPFILES: return SendMessage(hHexWnd, iMsg, wParam, lParam);

		case WM_SIZE:{
			SendMessage(hwndStatusBar, WM_SIZE, 0 , 0); //Moves status bar back to the bottom

			//--------------------------------------------
			// Set statusbar divisions.
			int statbarw;
			statbarw = LOWORD(lParam);
			// Allocate an array for holding the right edge coordinates.
			HLOCAL hloc = LocalAlloc (LHND, sizeof(int) * 3);
			int* lpParts = (int*) LocalLock(hloc);

			// Calculate the right edge coordinate for each part, and
			// copy the coordinates to the array.
			lpParts[0] = statbarw*4/6;
			lpParts[1] = statbarw*5/6;
			lpParts[2] = statbarw;

			// Tell the status window to create the window parts.

			SendMessage (hwndStatusBar, SB_SETPARTS, (WPARAM) 3, (LPARAM) lpParts);

			// Free the array, and return.
			LocalUnlock(hloc);
			LocalFree(hloc);

			RECT rect;
			GetClientRect(hwndStatusBar, &rect);

			MoveWindow(hHexWnd, 0, 0, LOWORD(lParam), HIWORD(lParam)-rect.bottom, TRUE);
			break;
		}

		case WM_NOTIFY:{
			//See if someone sent us invalid data
			try{
				//Attempt to dereference
				HWND tmp = ((NMHDR*)lParam)->hwndFrom;
			}
			catch(...){
				return 0;
			}
			hexwnd.notify((NMHDR*)lParam);
			return 0;
		}

		case WM_DESTROY:
			PostQuitMessage(0);
		return 0;

	}
	return DefWindowProc(hwnd, iMsg, wParam, lParam );
}

// The hex window procedure.
LRESULT CALLBACK HexWndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam )
{
	return hexwnd.OnWndMsg( hwnd, iMsg, wParam, lParam );
}
//============================================================================================
